/*-
 * Copyright (C)2007 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2007 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: block.c,v 1.7 2007/03/08 23:30:10 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/socket.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "pgsadump.h"
#include "block.h"

#include "time.cc"

static int
PGsaD_bool(PGsaDumpCtl *ctl, void *q, int z) {
    if  (*((char*)q)) printf("t"); else printf("f");
    return(EX_OK);
}

static int
PGsaD_bit(PGsaDumpCtl *ctl, void *q, int z) {
    int i, j;
    u_char *c;

    j = *((int*)q);
    c = (u_char*)q;
    c += 4;
    if  ((j - 4) < ((*((int*)c) + 7) >> 3)) {
        ifBLIN_QV1(ctl->flags) fprintf(stderr, "Byte L=%d bit L=%d", j, *((int*)c));
        return(EX_DATAERR);
    }
    j = *((int*)c);
    c += 4;
    for (i = 0; i < j; i++) {
        if  (c[i >> 3] & (0x80 >> (i & 7 ))) printf("1");
          else printf("0");
    }
    return(EX_OK);
}

static int
PGsaD_int2(PGsaDumpCtl *ctl, void *q, int z) {
    printf("%d", *(int16_t*)q);
    return(EX_OK);
}

static int
PGsaD_int4(PGsaDumpCtl *ctl, void *q, int z) {
    printf("%d", *(int32_t*)q);
    return(EX_OK);
}

static int
PGsaD_int8(PGsaDumpCtl *ctl, void *q, int z) {
    printf("%lld", *(int64_t*)q);
    return(EX_OK);
}

static int
PGsaD_oid(PGsaDumpCtl *ctl, void *q, int z) {
    printf("%u", *(u_int32_t*)q);
    return(EX_OK);
}

static int
PGsaD_float4(PGsaDumpCtl *ctl, void *q, int z) {
    float32_t num;
    int       infflag;

    num = *((float32_t*)q);
    infflag = isinf(num);
    if  (isnan(num)) printf("NaN");
    else if (infflag > 0) printf("Infinity");
    else if (infflag < 0) printf("-Infinity");
    else printf("%.*g", FLT_DIG, num);
    return(EX_OK);
}

static int
PGsaD_float8(PGsaDumpCtl *ctl, void *q, int z) {
    float64_t num;
    int       infflag;

    num = *((float64_t*)q);
    infflag = isinf(num);
    if  (isnan(num)) printf("NaN");
    else if (infflag > 0) printf("Infinity");
    else if (infflag < 0) printf("-Infinity");
    else printf("%.*g", DBL_DIG, num);
    return(EX_OK);
}

static int
PGsaD_numeric(PGsaDumpCtl *ctl, void *q, int z) {
    NumericData *num = q;
    int16_t      dig, d1;
    int          i, d;
    bool         putit;
    int          ndigits; /* # of digits in digits[] - can be 0!      */
    int          weight;  /* weight of first digit                    */
    int          dscale;  /* display scale                            */

    if  (NUMERIC_IS_NAN(num)) {
        printf("NaN");
        return(EX_OK);
    }
    if  ((ctl->nbase != 10) && (ctl->nbase != 100) && (ctl->nbase != 10000)) {
        errno = EAFNOSUPPORT;
        ifBLIN_QV1(ctl->flags) warnx("Invalid numeric: NBASE=%u", ctl->nbase);
        return(EX_DATAERR);
    }
    /*
     * Get the number in the variable format.
     *
     * Even if we didn't need to change format, we'd still need to copy the
     * value to have a modifiable copy for rounding.  set_var_from_num()
     * also guarantees there is extra digit space in case we produce a
     * carry out from rounding.
     */
    ndigits = (num->varlen - NUMERIC_HDRSZ) / PG_DIGITS(ctl->nbase);
    weight = num->n_weight;
    dscale = NUMERIC_DSCALE(num);
    if  (dscale < 0) dscale = 0;
    if  (NUMERIC_SIGN(num) == NUMERIC_NEG) printf("-");
    /*
     * Output all digits before the decimal point
     */
    if  (weight < 0) {
        d = weight + 1;
        printf("0");
    } else {
        for (d = 0; d <= weight; d++) {
            dig = (d < ndigits) ? DIGIT(ctl->nbase, num->n_data, d) : 0;
            /* In the first digit, suppress extra leading decimal zeroes */
            if  (NBASE(ctl->nbase, 10000)) {
                putit = (d > 0);
                MODULO(dig, d1, 1000);
                if  ((putit |= (d1 > 0))) printf("%01d", d1);
                MODULO(dig, d1, 100);
                if  ((putit |= (d1 > 0))) printf("%01d", d1);
                MODULO(dig, d1, 10);
                if  ((putit |= (d1 > 0))) printf("%01d", d1);
                printf("%01d", dig);
            } else if (NBASE(ctl->nbase, 100)) {
                MODULO(dig, d1, 10);
                if  (d1 > 0 || d > 0) printf("%01d", d1);
                printf("%01d", dig);
            } else if (NBASE(ctl->nbase, 10)) {
                printf("%01d", dig);
            } else {
                /* Should never be there */
                errno = EAFNOSUPPORT;
                ifBLIN_QV1(ctl->flags) warnx("Invalid#1 numeric: NBASE=%u", ctl->nbase);
                return(EX_DATAERR);
    }   }   }
    /*
     * If requested, output a decimal point and all the digits that follow
     * it.  We initially put out a multiple of DEC_DIGITS digits, then
     * truncate if needed.
     */
    if  (dscale > 0) {
        printf(".");
        for (i = 0; i < dscale; d++, i += DEC_DIGITS(ctl->nbase)) {
            dig = (d >= 0 && d < ndigits) ? DIGIT(ctl->nbase, num->n_data, d) : 0;
            if  (NBASE(ctl->nbase, 10000)) {
                MODULO(dig, d1, 1000);
                printf("%01d", d1);
                MODULO(dig, d1, 100);
                printf("%01d", d1);
                MODULO(dig, d1, 10);
                printf("%01d", d1);
                printf("%01d", dig);
            } else if (NBASE(ctl->nbase, 100)) {
                MODULO(dig, d1, 10);
                printf("%01d", d1);
                printf("%01d", dig);
            } else if (NBASE(ctl->nbase, 10)) {
                printf("%01d", dig);
            } else {
                /* Should never be there */
                errno = EAFNOSUPPORT;
                ifBLIN_QV1(ctl->flags) warnx("Invalid#2 numeric: NBASE=%u", ctl->nbase);
                return(EX_DATAERR);
    }   }   }
    return(EX_OK);
}

static int
PGsaD_char(PGsaDumpCtl *ctl, void *q, int z) {
#define CONTEXT(z) ((z) & (PGsaDumpAQ | PGsaDumpBA))
#define E0(i)      (e[CONTEXT(i)])
#define E1(i)      (e[CONTEXT(i) + 2])
#define SLS 0x01
#define ESC 0x02
#define BSL 0x04
#define OCT 0x08
#define ORG 0x10
#define SUB 0x20
    static const u_char t[] =
    {  0 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 , 'b','t','n','v', 'f','r', 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  2 , 2 , 3 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2
    ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2
    ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2
    ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  4 , 2 , 2 , 2
    ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2
    ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 2 ,  2 , 2 , 2 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1 ,  1 , 1 , 1 , 1
    };
    static const u_char a[4][5] =
    /*  /0       CTRL    reg     "        \  */
    { {SLS|OCT,     ORG, ORG,     ORG, ESC|BSL} /* text         */
    , {SLS|OCT,     ORG, ORG, ESC|ORG, ESC|BSL} /* quoted text  */
    , {ESC|OCT, ESC|OCT, ORG,     ORG, ESC|BSL} /* bytea        */
    , {ESC|OCT, ESC|OCT, ORG, SLS|ORG, ESC|BSL} /* quoted bytea */
    };
    static const char *e[] =
/* E0 */ {"\\", "\\"
/* E1 */ , "\\", "\\\\", "\\\\", "\\\\\\\\"
         };
    u_char     c = *(u_char*)q;
    u_char     s, tc = t[c];

    if  (z & PGsaDumpAL) printf("\"");
    if  ((z & PGsaDumpBA) && (tc > ' ')) tc = 1;
    if  (tc > ' ') s = SLS | SUB;
      else s = a[CONTEXT(z)][tc];
    if  (s & SLS) printf("%s", E0(z));
    if  (s & ESC) printf("%s", E1(z));
    if  (s & BSL) printf("%s", E1(z));
    if  (s & OCT) printf("%03o", c);
    if  (s & ORG) printf("%c", c);
    if  (s & SUB) printf("%c", t[c]);
    if  (z & PGsaDumpAL) printf("\"");
    return(EX_OK);
#undef SUB
#undef ORG
#undef OCT
#undef BSL
#undef ESC
#undef SLS
#undef E1
#undef E0
#undef CONTEXT
}

static int
PGsaD_name(PGsaDumpCtl *ctl, void *q, int z) {
    int        i;
    u_char    *c = (u_char*)q;
    int        e = EX_OK;

    if  (z & PGsaDumpAL) printf("\"");
    for (i = 0; i < NAMEDATALEN && !e && *c; i++) e = PGsaD_char(ctl, c++, (z & PGsaDumpAQ));
    if  (z & PGsaDumpAL) printf("\"");
    return(e);
}

static int
PGsaD_text(PGsaDumpCtl *ctl, void *q, int z) {
    int        i, j = *((int*)q) - sizeof(int32_t);
    u_char    *c = (u_char*)q + sizeof(int32_t);
    int        e = EX_OK;

    if  (z & PGsaDumpAL) printf("\"");
    for (i = 0; i < j && !e; i++) e = PGsaD_char(ctl, c++, (z & PGsaDumpAQ));
    if  (z & PGsaDumpAL) printf("\"");
    return(e);
}

static int
PGsaD_bytea(PGsaDumpCtl *ctl, void *q, int z) {
    int        i, j = *((int*)q) - sizeof(int32_t);
    u_char    *c = (u_char*)q + sizeof(int32_t);
    int        e = EX_OK;

    if  (z & PGsaDumpAL) printf("\"");
    for (i = 0; i < j && !e; i++) e = PGsaD_char(ctl, c++, (z & PGsaDumpAQ) | PGsaDumpBA);
    if  (z & PGsaDumpAL) printf("\"");
    return(e);
}

static int
PGsaD_abstime(PGsaDumpCtl *ctl, void *q, int z) {
    time_t     time = *((time_t*)q);
    struct tm *tx;
    int        e = EX_OK;
    Timestamp  fsec;

    switch (time) {
     case INVALID_ABSTIME:
        printf("invalid");
        break;
     case NOEND_ABSTIME:
        printf("infinity");
        break;
     case NOSTART_ABSTIME:
        printf("-infinity");
        break;
     default:
        if  (z & PGsaDumpAL) printf("\"");
        if  (ctl->flags & PGsaDumpTINT) fsec.i = 0; else fsec.d = 0;
        tx = gmtime(&time);
        tx->tm_year += 1900;
        tx->tm_mon += 1;
        tx->tm_isdst = -1;
        e = EncodeDateTime(ctl, tx, fsec);
        if  (e) ifBLIN_QV1(ctl->flags) warn("could not print abstime");
        if  (z & PGsaDumpAL) printf("\"");
    }
    return(e);
}

static int
PGsaD_reltime(PGsaDumpCtl *ctl, void *q, int z) {
    int32_t    time = *((int32_t*)q);
    struct tm  tt, *tm = &tt;
    int        e = EX_OK;
    Timestamp  fsec;

    if  (ctl->flags & PGsaDumpTINT) fsec.i = 0; else fsec.d = 0;
    MODULO(time, tt.tm_year, 31557600);
    MODULO(time, tt.tm_mon, 2592000);
    MODULO(time, tt.tm_mday, 86400);
    MODULO(time, tt.tm_hour, 3600);
    MODULO(time, tt.tm_min, 60);
    MODULO(time, tt.tm_sec, 1);
    if  (z & PGsaDumpAL) printf("\"");
    if  ((e = EncodeInterval(ctl, tm, fsec)) != 0) {
        ifBLIN_QV1(ctl->flags) warn("could not print reltime");
    }
    if  (z & PGsaDumpAL) printf("\"");
    return(e);
}

static int
PGsaD_interval(PGsaDumpCtl *ctl, void *q, int z) {
    Interval   *span = (Interval*)q;
    struct tm   tt, *tm = &tt;
    Timestamp   fsec;
    int e = EX_OK;

    interval2tm(ctl, *span, tm, &fsec);
    if  (z & PGsaDumpAL) printf("\"");
    if  ((e = EncodeInterval(ctl, tm, fsec)) != 0) {
        ifBLIN_QV1(ctl->flags) warn("could not print interval");
    }
    if  (z & PGsaDumpAL) printf("\"");
    return(e);
}

static int
PGsaD_date(PGsaDumpCtl *ctl, void *q, int z) {
    int32_t    date = *(int32_t*)q;
    struct tm  tt;
    int        e = EX_OK;

    j2date(date + POSTGRES_EPOCH_JDATE, &(tt.tm_year), &(tt.tm_mon), &(tt.tm_mday));
    if  ((e = EncodeDateOnly(ctl, &tt)) != EX_OK) {
        ifBLIN_QV1(ctl->flags) warn("could not print date");
    }
    return(e);
}

static int
PGsaD_time(PGsaDumpCtl *ctl, void *q, int z) {
    Timestamp  time = *((Timestamp*)q);
    struct tm  tt;
    Timestamp  fsec;
    int        e = EX_OK;

    time2tm(ctl, time, &tt, &fsec);
    if  ((e = EncodeTimeOnly(ctl, &tt, fsec, NULL)) != EX_OK) {
        ifBLIN_QV1(ctl->flags) warn("could not print time");
    }
    return(e);
}
 
static int
PGsaD_timetz(PGsaDumpCtl *ctl, void *q, int z) {
    TimeTzADT *time = ((TimeTzADT*)q);
    struct tm  tt;
    Timestamp  fsec;
    int        e = EX_OK;

    time2tm(ctl, time->time, &tt, &fsec);
    if  ((e = EncodeTimeOnly(ctl, &tt, fsec, &time->zone)) != EX_OK) {
        ifBLIN_QV1(ctl->flags) warn("could not print timetz");
    }
    return(e);
}

static int
PGsaD_timestamp(PGsaDumpCtl *ctl, void *q, int z) {
    Timestamp  timestamp = *((Timestamp*)q);
    struct tm  tt;
    Timestamp  fsec;
    int        e = EX_OK;

    if  (ctl->flags & PGsaDumpTINT) {
        if  (timestamp.i == DT_NOBEGIN) {
            printf("-infinity");
            return(EX_OK);
        }
        if  (timestamp.i == DT_NOEND) {
            printf("infinity");
            return(EX_OK);
        }
    } else {
        if  (timestamp.d == -HUGE_VAL) {
            printf("-infinity");
            return(EX_OK);
        }
        if  (timestamp.d == HUGE_VAL) {
            printf("infinity");
            return(EX_OK);
    }   }
    if  (z & PGsaDumpAL) printf("\"");
    if  ((e = timestamp2tm(ctl, timestamp, &tt, &fsec)) == EX_OK) e = EncodeDateTime(ctl, &tt, fsec);
    if  (e) ifBLIN_QV1(ctl->flags) warn("could not print timestamp");
    if  (z & PGsaDumpAL) printf("\"");
    return(e);
}

static int
PGsaD_macaddr(PGsaDumpCtl *ctl, void *q, int z) {
    macaddr   *addr = q;

    printf("%02x:%02x:%02x:%02x:%02x:%02x", addr->a, addr->b, addr->c, addr->d, addr->e, addr->f);
    return(EX_OK);
}

static int
PGsaD_inet(PGsaDumpCtl *ctl, void *q, int z) {
    inet      *src = q;
    u_char    *ip = ip_addr(src);
    int        e, i, bits = ip_bits(src);
    int        zero_s, zero_l, tmp_zero_s, tmp_zero_l;
    int        is_ipv4 = 0;

    e = EX_OK;
    switch (ip_family(src)) {
     case PGSQL_AF_INET:
        if  (bits < 0 || bits > 32) {
            errno = EINVAL;
            ifBLIN_QV1(ctl->flags) warnx("Invalid IP");
            e = EX_DATAERR;
        } else printf("%u.%u.%u.%u/%u", ip[0] & 0xFF, ip[1] & 0xFF, ip[2] & 0xFF, ip[3] & 0xFF, bits);
        break;
     case PGSQL_AF_INET6:
        if  (bits < 0 || bits > 128) {
            errno = EINVAL;
            ifBLIN_QV1(ctl->flags) warnx("Invalid IP");
            e = EX_DATAERR;
        } else {
        ;   zero_s = zero_l = tmp_zero_s = tmp_zero_l = 0;
        ;   for (i = 0; i < 16; i += 2) {                    /* Find the longest substring of zero's */
        ;       if  ((ip[i] | ip[i + 1]) == 0) {
        ;           if  (tmp_zero_l == 0) tmp_zero_s = i / 2;
        ;           tmp_zero_l++;
        ;       } else {
        ;           if  (tmp_zero_l && zero_l < tmp_zero_l) {
        ;               zero_s = tmp_zero_s;
        ;               zero_l = tmp_zero_l;
        ;               tmp_zero_l = 0;
        ;   }   }   }
        ;   if  (tmp_zero_l && zero_l < tmp_zero_l) {
        ;       zero_s = tmp_zero_s;
        ;       zero_l = tmp_zero_l;
        ;   }
        ;   if  (  zero_l != 8
                && zero_s == 0
                && (  (zero_l == 6)
                   || (  (zero_l == 5 && ip[10] == 0xff && ip[11] == 0xff)
                      || ((zero_l == 7 && ip[14] != 0 && ip[15] != 1))
                )  )  ) is_ipv4 = 1;

        ;   for (i = 0; i < 8; i++, ip += 2) {
        ;       if  (zero_l != 0 && i >= zero_s && i < zero_s + zero_l) {
        ;           if  (i == zero_s) printf(":");
        ;           if  (i == 7) printf(":");
        ;       } else if (is_ipv4 && i > 5) {
        ;           printf("%c%u.%u", (i == 6) ? ':' : '.', ip[0], ip[1]);
        ;       } else {
        ;           if  (i) printf(":");
        ;           printf("%x", *ip * 256 + ip[1]);
        ;       }
        ;   }
        ;   printf("/%u", bits);
        }
        break;
     default:
        errno = EAFNOSUPPORT;
        e = EX_DATAERR;
    }
    ifBLIN_QV1(ctl->flags) if (e) warnx("Invalid IP");
    return(e);
}

static int
PGsaD_array( PGsaDumpCtl *ctl, int *dim, int ndim, void *d, u_int32_t type) {
    u_int32_t  al = PGsaDumpCollection[type].align;
    ssize_t    len = 0;
    int        e = EX_OK;
    int        i;

    if  (al > ctl->alignmax) al = ctl->alignmax;
    al--;
    e = 0;
    ndim--;
    printf("{");
    for (i = 0; i < *dim && !e; i++) {
        if  (ndim) {
            if  (i) printf(",");
            e = PGsaD_array(ctl, dim + 1, ndim, d, type);
        } else {
            if  (i) printf("%s", PGsaDumpCollection[type].delim);
            d = (void*)(((u_int32_t)d + al) & ~al);
            len = PGsaDumpCollection[type].len;
            if  (len < 0) len = *((u_int32_t*)d);
            e = (*(PGsaDumpCollection[type].conv))(ctl, d, PGsaDumpAQ | PGsaDumpAL);
            d = (char*)d + len;
    }   }
    printf("}");
    return(e);
}

PGsaDumpCollection_t PGsaDumpCollection[] =
{{PGsaD_bool     ,  1, 1, ",", "bool"     }
,{PGsaD_bit      , -1, 4, ",", "bit"      }
,{PGsaD_int2     ,  2, 2, ",", "int2"     }
,{PGsaD_int4     ,  4, 4, ",", "int4"     }
,{PGsaD_int8     ,  8, 8, ",", "int8"     }
,{PGsaD_oid      ,  4, 4, ",", "oid"      }
,{PGsaD_float4   ,  4, 4, ",", "float4"   }
,{PGsaD_float8   ,  8, 8, ",", "float8"   }
,{PGsaD_numeric  , -1, 4, ",", "numeric"  }
,{PGsaD_char     ,  1, 1, ",", "char"     }
,{PGsaD_name     , 64, 4, ",", "name"     }
,{PGsaD_text     , -1, 4, ",", "text"     }
,{PGsaD_bytea    , -1, 4, ",", "bytea"    }
,{PGsaD_abstime  ,  4, 4, ",", "abstime"  }
,{PGsaD_reltime  ,  4, 4, ",", "reltime"  }
,{PGsaD_interval , 12, 8, ",", "interval" }
,{PGsaD_date     ,  4, 4, ",", "date"     }
,{PGsaD_time     ,  8, 8, ",", "time"     }
,{PGsaD_timetz   , 12, 8, ",", "timetz"   }
,{PGsaD_timestamp,  8, 8, ",", "timestamp"}
,{PGsaD_macaddr  ,  6, 4, ",", "macaddr"  }
,{PGsaD_inet     , -1, 4, ",", "inet"     }
,{NULL           ,  0, 0, NULL, NULL      }
};

int
PGsaDumpBlock(PGsaDumpCtl *ctl, u_int32_t pgn, void *pp) {
    PageHeaderData *p = (PageHeaderData*)pp;
    HeapTupleHeaderData *t;
    OffsetNumber n;
    ItemIdData *i;
    int j;
    int c; /* current print string is not empty */
    int e; /* error code */
    void *q;

    e = 0;
    ifBLIN_QV4(ctl->flags)
        fprintf(stderr, "+PGsaDumpBlock =%04X(%04X) N%d\n", ctl->atrterm, ctl->atrmask, ctl->atrmax);
    if  (!PageIsEmpty(p)) {
        if  (PageGetPageLayoutVersion(p) != PG_PAGE_LAYOUT_VERSION) {
            errno = EINVAL;
            ifBLIN_QV1(ctl->flags) warnx("PG_PAGE_LAYOUT_VERSION=%d", PageGetPageLayoutVersion(p));
            if  (!(ctl->flags & PGsaDumpXVER)) return(EX_DATAERR);
        }
        if  (PageGetPageSize(p) != ctl->blksz) {
            errno = EINVAL;
            ifBLIN_QV1(ctl->flags) warnx("PageSize=%d", PageGetPageSize(p));
            if  (!(ctl->flags & PGsaDumpXBSZ)) return(EX_DATAERR);
        }
        for (n = FirstOffsetNumber; n <= PageGetMaxOffsetNumber(p) && !e; n++) {
            i = PageGetItemId(p, n);
            t = (HeapTupleHeaderData*)PageGetItem(p, i);
            if  (!(  (ctl->flags & PGsaDumpETID)
                  || (  (pgn == ((t->t_ctid.ip_blkid.bi_hi << 16) |  t->t_ctid.ip_blkid.bi_lo))
                     && (n == t->t_ctid.ip_posid)
                ) )  ) continue;
            if  (!((ctl->flags & PGsaDumpLNUS) || ItemIdIsUsed(i))) continue;
            if  (!((ctl->flags & PGsaDumpLDEL) || !ItemIdDeleted(i))) continue;
            if  (t->t_natts < ctl->atrmin) continue;
            if  (t->t_natts > ctl->atrmax) continue;
            if  ((t->t_infomask ^ ctl->atrterm) & ctl->atrmask) continue;
            c = 0;
            if  (ctl->flags & PGsaDumpRTID) {
                printf(c ? "\t(%u,%d)" : "(%u,%d)", pgn, n);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpNATR) {
                printf(c ? "\t%u" : "%u", t->t_natts);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpINFO) {
                printf(c ? "\t%04X" : "%04X", t->t_infomask);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpCMAX) {
                printf(c ? "\t%u" : "%u", t->t_field3.t_cmax);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpXMAX) {
                printf( c ? "\t%u" : "%u"
                      , (t->t_infomask & HEAP_XMAX_IS_XMIN) ? t->t_xmin : t->t_field2.t_xmax
                      );
                c = 1;
            }
            if  (ctl->flags & PGsaDumpCMIN) {
                printf(c ? "\t%u" : "%u", t->t_field2.t_cmin);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpXMIN) {
                printf(c ? "\t%u" : "%u", t->t_xmin);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpOID) {
                if  (t->t_infomask & HEAP_HASOID) {
                    printf( c ? "\t%u" : "%u"
                          , *((Oid*)((char*)t + t->t_hoff - sizeof(Oid)))
                          );
                } else {
                    printf(c ? "\t%u" : "%u", InvalidOid);
                }
                c = 1;
            }
            if  (ctl->flags & PGsaDumpCTID) {
                printf( c ? "\t(%u,%d)" : "(%u,%d)"
                      , (t->t_ctid.ip_blkid.bi_hi << 16) |  t->t_ctid.ip_blkid.bi_lo
                      , t->t_ctid.ip_posid
                      );
                c = 1;
            }
            if  (ctl->flags & PGsaDumpNULS) {
                if  (c) printf("\t");
                for (j = 0; j < BITMAPLEN(t->t_natts); j++) printf("%02X", t->t_bits[j]);
                c = 1;
            }
            q = (char *)t + t->t_hoff; /* tuples begin */
            for (j = 0; j < ctl->atrs[0] && !e; j++) {
            ;   if  (!(ctl->flags & PGsaDumpXATR || j < t->t_natts)) break;
            ;   if  (c) printf("\t");
            ;   if  (  ((t->t_infomask & HEAP_HASNULL) && !(t->t_bits[j >> 3] & (1 << (j & 7))))
                    || !(j < t->t_natts)
                    ) {
                    printf("\\N");
            ;   } else {
                    size_t len;
                    u_int32_t type = ctl->atrs[j + 1] & 0x7FFF;
                    u_int32_t array = ctl->atrs[j + 1] & 0x8000;
                    u_int32_t al = PGsaDumpCollection[type].align;
            ;       if  (array || al > ctl->alignmax) al = ctl->alignmax;
            ;       al--;
            ;       ifBLIN_QV3(ctl->flags)
                        fprintf( stderr, "p=%08X t=%08X q=%08X a=%08X"
                               , (u_int32_t)p, (u_int32_t)t, (u_int32_t)q, al
                               );
            ;       q = (void*)(((u_int32_t)q + al) & ~al);
            ;       if  (array || (PGsaDumpCollection[type].len < 0)) len = *((u_int32_t*)q);
                      else len = PGsaDumpCollection[type].len;
            ;       ifBLIN_QV3(ctl->flags)
                        fprintf( stderr, " q=%08X N=%d L=%d%s\n"
                               , (u_int32_t)q, j + 1, len, array ? " A" : ""
                               );
            ;       if  (  ((char*)p + ctl->blksz) < (char*)q
                        || ((char*)p + ctl->blksz) < ((char*)q + len)
                        ) {
            ;           ifBLIN_QV2(ctl->flags) fprintf(stderr, "   \n");
            ;           if  (!(ctl->flags & PGsaDumpXTUP)) e = EX_DATAERR;
            ;           continue;
            ;       }
            ;       if  (ItemIdGetLength(i) < len) {
            ;           /* XXXX     ,   */
            ;           ifBLIN_QV2(ctl->flags) fprintf(stderr, "   \n");
            ;           if  (!(ctl->flags & PGsaDumpXTUP)) e = EX_DATAERR;
            ;           continue;
            ;       }
            ;       if  (array) {
                        int k, ndim, *dim;
                        void *d;
                        u_int64_t osize;

            ;           ndim = ((ArrayType*)q)->ndim;
            ;           if  (ndim > MAXDIM) {
            ;               ifBLIN_QV2(ctl->flags) fprintf(stderr, "%d- \n", ndim);
            ;               if  (!(ctl->flags & PGsaDumpXTUP)) e = EX_DATAERR;
            ;               continue;
            ;           }
            ;           dim = (int*)(((char*)q) + sizeof(ArrayType));
            ;           osize = 1;
            ;           for (k = 0; k < ndim; k++) {
            ;               osize *= dim[k];
            ;               if  (osize > len) {
            ;                   osize = 0;
            ;                   break;
            ;           }   }
            ;           if  (!osize) {
            ;               ifBLIN_QV2(ctl->flags) fprintf(stderr, " \n");
            ;               if  (!(ctl->flags & PGsaDumpXTUP)) e = EX_DATAERR;
            ;               continue;
            ;           }
            ;           d = (void*)(2 * sizeof(int) * ndim + sizeof(ArrayType) + (char *)q);
            ;           d = (void*)(((u_int32_t)d + ctl->alignmax - 1) & ~(ctl->alignmax - 1));
            ;           e = PGsaD_array(ctl, dim, ndim, d, type);
            ;       } else {
            ;           e = (*(PGsaDumpCollection[type].conv))(ctl, q, 0);
            ;       }
            ;       if  (ctl->flags & PGsaDumpXCOL) e = 0;
            ;       q = (char*)q + len;
            ;   }
            ;   c = 1;
            }
            if  (c) printf("\n");
    }   }
    ifBLIN_QV4(ctl->flags) fprintf(stderr, "-PGsaDumpBlock %d\n", e);
    return(e);
}
